#!/usr/bin/env python
import argparse
import os
import sys
import time

wd=os.path.dirname(os.path.realpath(__file__))
sys.path.insert(0, '{}/src/'.format(wd))
import TIDDIT_calling

version = "2.12.1"
parser = argparse.ArgumentParser("""TIDDIT-{}""".format(version),add_help=False)
parser.add_argument('--sv'       , help="call structural variation", required=False, action="store_true")
parser.add_argument('--cov'        , help="generate a coverage bed file", required=False, action="store_true")
args, unknown = parser.parse_known_args()

if args.sv:
	parser = argparse.ArgumentParser("""TIDDIT --sv --bam inputfile [-o prefix] --ref ref.fasta""")
	parser.add_argument('--sv'       , help="call structural variation", required=False, action="store_true")
	parser.add_argument('--bam', type=str,required=True, help="coordinate sorted bam file(required)")
	parser.add_argument('-o', type=str,default="output", help="output prefix(default=output)")
	parser.add_argument('-i', type=int, help="paired reads maximum allowed insert size. Pairs aligning on the same chr at a distance higher than this are considered candidates for SV (default= 99.9th percentile of insert size)")
	parser.add_argument('-d', type=str,help="expected reads orientations, possible values \"innie\" (-> <-) or \"outtie\" (<- ->). Default: major orientation within the dataset")
	parser.add_argument('-p', type=int,default=3, help="Minimum number of supporting pairs in order to call a variation event (default 3)")
	parser.add_argument('-r', type=int,default=3, help="Minimum number of supporting split reads to call a small variant (default 3)")
	parser.add_argument('-q', type=int,default=5, help="Minimum mapping quality to consider an alignment (default 5)")
	parser.add_argument('-Q', type=int,default=20, help="Minimum regional mapping quality (default 20)")
	parser.add_argument('-n', type=int,default=2, help="the ploidy of the organism,(default = 2)")
	parser.add_argument('-e', type=int, help="clustering distance  parameter, discordant pairs closer than this distance are considered to belong to the same variant(default = sqrt(insert-size*2)*12)")
	parser.add_argument('-l', type=int,default=3, help="min-pts parameter (default=3),must be set >= 2")
	parser.add_argument('-s', type=int,default=25000000, help="Number of reads to sample when computing library statistics(default=25000000)")
	parser.add_argument('-z', type=int,default=100, help="minimum variant size (default=100), variants smaller than this will not be printed ( z < 10 is not recomended)")
	parser.add_argument('--force_ploidy',action="store_true", help="force the ploidy to be set to -n across the entire genome (i.e skip coverage normalisation of chromosomes)")
	parser.add_argument('--no_cluster',action="store_true", help="Run only the TIDDIT signal extraction")
	parser.add_argument('--debug',action="store_true", help="rerun the tiddit clustering procedure")
	parser.add_argument('--n_mask',type=float,default=0.5, help="exclude regions from coverage calculation if they contain more than this fraction of N (default = 0.5)")
	parser.add_argument('--ref', type=str, help="reference fasta, used for GC correction and for reading cram")
	parser.add_argument('--p_ratio', type=float,default=0.2, help="minimum discordant pair/normal pair ratio at the breakpoint junction(default=0.2)")
	parser.add_argument('--r_ratio', type=float,default=0.1, help="minimum split read/coverage ratio at the breakpoint junction(default=0.1)")

	args= parser.parse_args()
	args.wd=os.path.dirname(os.path.realpath(__file__))
	if args.l < 2:
		print ("error, too low --l value!")
		quit()
	if args.ref:
		if not os.path.isfile(args.ref):
			print ("error,  could not find the reference file")
			quit()

	if not args.ref and args.bam.endswith(".cram"):
		print("error, reference fasta is required when using cram input")
		quit()

	if not (args.bam.endswith(".bam") or args.bam.endswith(".cram")) and not "/dev/" in args.bam:
		print ("error, the input file is not a bam file, make sure that the file extension is .bam or .cram")
		quit()

	if not os.path.isfile(args.bam) and not "/dev/" in args.bam:
		print ("error,  could not find the bam file")
		quit()

	if not os.path.isfile("{}/bin/TIDDIT".format(args.wd)):
		print ("error,  could not find the TIDDIT executable file, try rerun the INSTALL.sh script")
		quit()

	if not args.bam.endswith(".cram"):
		command_str="{}/bin/TIDDIT --sv -b {} -o {} -p {} -r {} -q {} -n {} -s {}".format(args.wd,args.bam,args.o,args.p,args.r,args.q,args.n,args.s)
	else:
		command_str="samtools view -hbu {} -T {} | {}/bin/TIDDIT --sv -b /dev/stdin -o {} -p {} -r {} -q {} -n {} -s {}".format(args.bam,args.ref,args.wd,args.o,args.p,args.r,args.q,args.n,args.s)
	

	if args.i:
		command_str += " -i {}".format(args.i)
	if args.d:
		command_str += " -d {}".format(args.d)

	if not args.debug:
		os.system(command_str)

		if args.ref:
			t=time.time()
			print ("Generating GC wig file")
			if args.ref.endswith(".gz"):
				os.system("zcat {} | {}/bin/TIDDIT --gc -z 50 -o {}".format(args.ref,args.wd,args.o))
			else:
				os.system("cat {} | {}/bin/TIDDIT --gc -z 50 -o {}".format(args.ref,args.wd,args.o))
			print ("Constructed GC wig in {} sec".format(time.time()-t))

	if not args.no_cluster:
		TIDDIT_calling.cluster(args)

elif args.cov:
	parser = argparse.ArgumentParser("""TIDDIT --cov --bam inputfile [-o prefix]""")
	parser.add_argument('--cov'        , help="generate a coverage bed/wig file", required=False, action="store_true")
	parser.add_argument('--bam', type=str,required=True, help="coordinate sorted bam file(required)")
	parser.add_argument('-o', type=str,default="output", help="output prefix(default=output)")
	parser.add_argument('-z', type=int,default=500, help="use bins of specified size(default = 500bp) to measure the coverage of the entire bam file, set output to stdout to print to stdout")
	parser.add_argument('-w'        , help="generate wig instead of bed", required=False, action="store_true")
	parser.add_argument('-u'        , help="skip per bin mapping quality", required=False, action="store_true")
	parser.add_argument('--ref', type=str, help="reference fasta, used for GC correction and for reading cram")
	args= parser.parse_args()
	args.wd=os.path.dirname(os.path.realpath(__file__))

	if args.ref:
		if not os.path.isfile(args.ref):
			print ("error,  could not find the reference file")
			quit()

	if not args.bam.endswith(".cram"):
		command="{}/bin/TIDDIT --cov -b {} -o {} -z {}".format(args.wd,args.bam,args.o,args.z)
	else:
		if not args.ref:
			print("error, missing reference sequence!")
			quit()
		command="samtools view -hbu {} -T {} | {}/bin/TIDDIT --cov -b /dev/stdin -o {} -z {}".format(args.bam,args.ref,args.wd,args.o,args.z)

	if args.w:
		command += " -w"

	if args.u:
		command += " -u"

	if not os.path.isfile(args.bam):
		print ("error,  could not find the bam file")
		quit()
	if not os.path.isfile("{}/bin/TIDDIT".format(args.wd)):
		print ("error,  could not find the TIDDIT executable file, try rerun the INSTALL.sh script")
		quit()
	os.system(command)

else:
	parser.print_help()



import itertools
import sqlite3
import math
import numpy

#The functions of this script are used to read the input signals, and to create the database of signals

#extract the vcf header of the signals tab file
def find_contigs(header):
	chromosomes=[]
	library_stats={}
	chromosome_len={}
	for line in header.split("\n"):
		if "##contig=<ID=" in line:
			chromosomes.append(line.split("##contig=<ID=")[-1].split(",")[0])
			chromosome_len[chromosomes[-1]]=int(line.split("length=")[-1].split(">")[0])
		if "##LibraryStats=" in line:
			stats=line.strip().split(" ")
			library_stats={"Coverage":float(stats[1].split("=")[-1]),"ReadLength":int(stats[2].split("=")[-1]),"MeanInsertSize":int(stats[3].split("=")[-1]),"STDInsertSize":int(stats[4].split("=")[-1]),"Orientation":stats[5].split("=")[-1]}
	return (chromosomes,library_stats,chromosome_len)

def sample(args,coverage_data,span_data):
	n=2000
	bridges=[]
	bridges_reads=[]
	bin_size=50

	for chromosome in coverage_data:

		if bin_size*len(coverage_data[chromosome][:,0]) < 200000:
			continue
		index=numpy.random.randint(low=1000, high=len(coverage_data[chromosome][:,0])-1000,size=n)
		for i in range(0,n):
			
			coverage=coverage_data[chromosome][index[i],0]
			if coverage == 0:
				continue

			bridging=span_data[chromosome][index[i],0]
			bridging_reads=span_data[chromosome][index[i],1]
			bridges_reads.append(bridging_reads/coverage)
			bridges.append(bridging/coverage)

	percentile_list=numpy.arange(0,101,1)
	
	if len(bridges):
		percentiles_disc=numpy.percentile(bridges,percentile_list)
		percentiles_splits=numpy.percentile(bridges_reads,percentile_list)
	else:
		print ("Warning - to few reads in the bam file, skipping permutation tests")
		percentiles_disc=[]
		percentiles_splits=[]

	#for i in range(0,len(percentiles_disc)):
	#	print "{} {} {}".format(i,percentiles_disc[i], percentiles_splits[i])

	return (percentiles_disc,percentiles_splits)

#Read the signals tab file
def signals(args,coverage_data):
	signal_data=[]
	header=""
	first_signal=True
	read_list={}
	n_reads=0
	for line in open(args.o+".signals.tab"):
		if line[0] == "#":
			header += line
			continue

		if first_signal:
			chromosomes,library_stats,chromosome_len=find_contigs(header)
			first_signal=False

		content=line.strip().split()

		read_name=content[0]
		if read_name in read_list:
			name=read_list[read_name]
		else:
			name=n_reads
			read_list[read_name]=n_reads
			n_reads+=1

		chrA=content[1]
		posA=int(content[2])
		if content[3] == "+":
			orientationA=1
		else:
			orientationA=0

		cigarA=content[4]
		if ("S" in cigarA or "H" in cigarA) and not cigarA == "NA":
			deletions,insertions,length,clip_after,aligned_range=read_cigar(cigarA)
			if clip_after:
				posA+=length
		else:
			if "M" in cigarA:
				deletions,insertions,length,clip_after,aligned_range=read_cigar(cigarA)
			else:
				length=library_stats["ReadLength"]

			if library_stats["Orientation"] == "innie":			
				if orientationA:
					posA+=length-1
			else:
				if not orientationA:
					posA+=length-1

		if posA >= chromosome_len[chrA]:
			posA=chromosome_len[chrA]-1

		qualA=int(content[5])

		chrB=content[6]
		posB=int(content[7])
		if content[8] == "+":
			orientationB=1
		else:
			orientationB=0
		cigarB=content[9]
		qualB=int(content[10])

		if ("S" in cigarB or "H" in cigarB) and not cigarB == "NA":
			deletions,insertions,length,clip_after,aligned_range=read_cigar(cigarB)
			if clip_after:
				posB+=length
		else:
			if "M" in cigarB:
				deletions,insertions,length,clip_after,aligned_range=read_cigar(cigarB)
			else:
				length=library_stats["ReadLength"]

			if library_stats["Orientation"] == "innie":			
				if orientationB:
					posB+=length-1
			else:
				if not orientationB:
					posB+=length-1

		if posB >= chromosome_len[chrB]:
			posB=chromosome_len[chrB]-1

		resolution=int(content[-1])
		

		if chrA > chrB or (posB < posA and chrA == chrB):
			signal_data.append([chrB,chrA,posB,posA,orientationB,orientationA,qualB,qualA,cigarB,cigarA,resolution,name])
		else:
			signal_data.append([chrA,chrB,posA,posB,orientationA,orientationB,qualA,qualB,cigarA,cigarB,resolution,name])
		

		if len (signal_data) > 1000000:
			args.c.executemany('INSERT INTO TIDDITcall VALUES (?,?,?,?,?,?,?,?,?,?,?,?)',signal_data)  
			signal_data=[]

		idx_b=int(math.floor(posB/50.0))
		if idx_b >= len(coverage_data[chrB]):
			idx_b =len(coverage_data[chrB])-1
		coverage_data[chrB][idx_b,2]+=1
                
		idx_a=int(math.floor(posA/50.0))
		if idx_a >= len(coverage_data[chrA]):
			idx_a=len(coverage_data[chrA])-1
		coverage_data[chrA][idx_a,2]+=1


	if len(signal_data):
		args.c.executemany('INSERT INTO TIDDITcall VALUES (?,?,?,?,?,?,?,?,?,?,?,?)',signal_data)  

	#If no signals were found, parse and return the header
	if first_signal:
		chromosomes,library_stats,chromosome_len=find_contigs(header)

	return(header,chromosomes,library_stats)

#analyse the cigar operation of the read
def read_cigar(cigar):
	deletions=0
	insertions=0
	SC = ["".join(x) for _, x in itertools.groupby(cigar, key=str.isdigit)]
	length=0
	first=True
	clip_after=True

	aligned_range=[]
	current_pos=1
	for i in range(0,len(SC)/2):
		if first and SC[i*2+1] == "M":
			first = False
		elif first and SC[i*2+1] == "S":
			first = False
			clip_after=False

		if SC[i*2+1] == "M":
			length += int( SC[i*2] )
			bases=list(range(0,int( SC[i*2] )))
			for j in range(0,len(bases)):
				bases[j] += current_pos

			aligned_range += bases
			current_pos += int( SC[i*2] )
		elif SC[i*2+1] == "I":
			insertions+=1
			bases=list(range(0,int( SC[i*2] )))
			for j in range(0,len(bases)):
				bases[j] += current_pos
			aligned_range += bases

			current_pos += int( SC[i*2] )
		elif SC[i*2+1] == "D":
			length += int( SC[i*2] )
			deletions +=1
		else:
			current_pos += int( SC[i*2] )

	return deletions,insertions,length,clip_after,aligned_range



import numpy

#compute the filters
def fetch_filter(chrA,chrB,candidate,args,library_stats):
	filt="PASS"

	#Less than the expected number of signals
	r_a=0
	if library_stats["chr_cov"][chrA]:
		r_a=candidate["covA"]/library_stats["chr_cov"][chrA]
	r_b=0
	if library_stats["chr_cov"][chrB]:
		r_b=candidate["covB"]/library_stats["chr_cov"][chrB]


	ratio_scaling=min([ r_b, r_a ])

	if ratio_scaling > 2:
		ratio_scaling=2
	elif ratio_scaling < 1:
		ratio_scaling=1

	if chrA == chrB and library_stats["ploidies"][chrA] > 10 and candidate["ratio"] <= 0.05:
		filt = "BelowExpectedLinks"
	elif candidate["ratio"]*ratio_scaling <= args.p_ratio and candidate["discs"] > candidate["splits"]:
		filt = "BelowExpectedLinks"
	elif candidate["ratio"]*ratio_scaling <= args.r_ratio and candidate["discs"] < candidate["splits"]:
		filt = "BelowExpectedLinks"

	#The ploidy of this contig is 0, hence there shoud be no variant here
	if library_stats["ploidies"][chrA] == 0 or library_stats["ploidies"][chrB] == 0:
		return("Ploidy")

	#coverage is too high
	if candidate["covA"] >= library_stats["chr_cov"][chrA]*(library_stats["ploidies"][chrA]*4+library_stats["ploidies"][chrA]) or candidate["covB"] >= library_stats["chr_cov"][chrB]*(library_stats["ploidies"][chrB]*4+library_stats["ploidies"][chrA]):
		filt = "UnexpectedCoverage"
	elif candidate["discsA"] > (candidate["discs"]+candidate["splits"])*(library_stats["ploidies"][chrA]*2) or candidate["discsB"] > (candidate["discs"]+candidate["splits"])*(library_stats["ploidies"][chrA]*2):
		filt= "FewLinks"
	elif chrA == chrB and candidate["max_A"] > candidate["min_B"]:
		filt = "Smear"
	elif chrA != chrB or (abs(candidate["posA"]-candidate["posB"]) > library_stats["MeanInsertSize"]+3*library_stats["STDInsertSize"] ):
		if not candidate["discs"] or candidate["discs"]*4 <  candidate["splits"] or candidate["discs"] <= args.p/2:
			filt= "SplitsVSDiscs"
	#fewer links than expected
	if chrA == chrB and (abs(candidate["max_A"]-candidate["min_B"]) < args.z):
		filt="MinSize"

	return(filt)

#determine the variant type
def fetch_variant_type(chrA,chrB,candidate,args,library_stats,disc_ratio,split_ratio):
	variant_type="SVTYPE=BND"
	var="N[{}:{}[".format(chrB,candidate["posB"])
	if not library_stats["ploidies"][chrA] and chrA == chrB:
		variant_type="SVTYPE=DUP"
		var="<DUP>"		

	ploidy_factor= 0.6
	if chrA == chrB and library_stats["ploidies"][chrA]:
		ploidy=library_stats["ploidies"][chrA]
		if ploidy > 10:
			if candidate["discs"] and abs(candidate["covM"]/library_stats["chr_cov"][chrA]-1) < 0.05:
				if candidate["FF"] + candidate["RR"] > candidate["RF"] + candidate["FR"]:
					variant_type="SVTYPE=INV"
					var="<INV>"
			elif not candidate["discs"] and abs(candidate["covM"]/library_stats["chr_cov"][chrA]-1) < 0.05:
				if candidate["splitsINV"] > candidate["splits"]-candidate["splitsINV"]:
					variant_type="SVTYPE=INV"
					var="<INV>"
			elif candidate["covM"]/library_stats["chr_cov"][chrA]-1 > 0.05:
				variant_type="SVTYPE=DUP"
				var="<DUP>"
			elif candidate["covM"]/library_stats["chr_cov"][chrA]-1 < -0.05:
				variant_type="SVTYPE=DEL"
				var="<DEL>"		

		elif candidate["discs"]:
			if candidate["FF"] + candidate["RR"] > candidate["RF"] + candidate["FR"]:
				variant_type="SVTYPE=INV"
				var="<INV>"
			elif library_stats["Orientation"] == "innie":
				if candidate["covM"]/library_stats["chr_cov"][chrA] > (ploidy+ploidy_factor)/float(ploidy):
					variant_type="SVTYPE=DUP"
					var="<DUP>"
					if candidate["RF"] > candidate["FR"]: 
						variant_type="SVTYPE=TDUP"
						var="<TDUP>"
				elif candidate["covM"]/library_stats["chr_cov"][chrA] < (ploidy-ploidy_factor)/float(ploidy):
					variant_type="SVTYPE=DEL"
					var="<DEL>"	

			else:
				if candidate["covM"]/library_stats["chr_cov"][chrA] > (ploidy+ploidy_factor)/float(ploidy):
					variant_type="SVTYPE=DUP"
					var="<DUP>"
					if candidate["RF"] < candidate["FR"]: 
						variant_type="SVTYPE=TDUP"
						var="<TDUP>"

				elif candidate["covM"]/library_stats["chr_cov"][chrA] < (ploidy-ploidy_factor)/float(ploidy):
					variant_type="SVTYPE=DEL"
					var="<DEL>"
		else:
			if candidate["splitsINV"] > candidate["splits"]-candidate["splitsINV"]:
				variant_type="SVTYPE=INV"
				var="<INV>"
			elif candidate["covM"]/library_stats["chr_cov"][chrA] >(ploidy+ploidy_factor)/float(ploidy):
					variant_type="SVTYPE=DUP"
					var="<DUP>"
			elif candidate["covM"]/library_stats["chr_cov"][chrA] < (ploidy-ploidy_factor)/float(ploidy):
					variant_type="SVTYPE=DEL"
					var="<DEL>"		


	if candidate["discs"] > candidate["splits"]:

		if disc_ratio >= 0.9:
			GT="1/1"
		else:
			GT="0/1"
	else:
		if split_ratio >= 0.9:
			GT="1/1"
		else:
			GT="0/1"

	if ("DUP" in var or var == "<DEL>") and library_stats["chr_cov"][chrA] != 0:
		if var == "<DEL>" and candidate["covM"]/library_stats["chr_cov"][chrA] < 0.1:
			GT="1/1"
		elif "DUP" in var and candidate["covM"]/library_stats["chr_cov"][chrA] > 1.8: 
			GT="1/1"
		else:
			GT="0/1"

	return(var,variant_type,GT)



import numpy

#The functions in this script are used to read and proccess the coverage tab file, as well as to perform GC correction

#read the coverage tab file
def coverage(args):
	chromosome_size={}
	for line in open(args.o+".wig"):
		if "chrom=" in line:
			chromosome=line.split("chrom=")[-1].split()[0]
			continue
		elif "quality" in line:
			break
		elif "descrip" in line:
			continue

		if not chromosome in chromosome_size:
			chromosome_size[chromosome]=0
		chromosome_size[chromosome]+=1
	
	coverage_data={}
	for chromosome in chromosome_size:
		coverage_data[chromosome]=numpy.zeros( (chromosome_size[chromosome],3), dtype=numpy.float32 )

	span_coverage={}
	for chromosome in chromosome_size:
		span_coverage[chromosome]=numpy.zeros( (chromosome_size[chromosome],2), dtype=numpy.uint32 )


	mapq=False
	spanning_pair=False
	spanning_read=False
	for line in open(args.o+".wig"):
		if "name=\"Coverage\"" in line:
			continue
		elif "chrom=" in line:
			chromosome=line.split("chrom=")[-1].split()[0]
			chromosome_index=0
			continue
		elif "quality" in line:
			mapq=True
			continue
		elif "SpanPairs" in line:
			spanning_pair=True
			continue
		elif "SpanReads" in line:
			spanning_read=True
			continue

		content=line.strip()	
		if mapq and not spanning_read and not spanning_pair:
			coverage_data[chromosome][chromosome_index][1]=float(content)
		elif mapq and spanning_pair and not spanning_read:
			span_coverage[chromosome][chromosome_index][0]=int(content)
		elif mapq and spanning_pair and  spanning_read:
			span_coverage[chromosome][chromosome_index][1]=int(content)
		else:
			coverage_data[chromosome][chromosome_index][0]=float(content)

		chromosome_index+=1

	return(coverage_data,span_coverage)

#estimate the ploidy of each chromosome
def determine_ploidy(args,chromosomes,coverage_data,Ncontent,library_stats):
	library_stats["chr_cov"]={}
	ploidies={}
	avg_coverage=[]

	cov=numpy.array([],dtype=numpy.float32)
	for chromosome in chromosomes:		
		try:
			if args.ref:
				N_count=Ncontent[chromosome]
				chr_cov=coverage_data[chromosome][numpy.where( (N_count > 0) & (coverage_data[chromosome][:,1] > args.Q)  ),0][0]
			else:
				chr_cov=coverage_data[chromosome][numpy.where( coverage_data[chromosome][:,1] > args.Q  ),0][0]

			if len(chr_cov):
				chromosomal_average=numpy.median(chr_cov)
				cov = numpy.append(cov,chr_cov)
			else:
				chromosomal_average=0
			library_stats["chr_cov"][chromosome]=chromosomal_average

		except:
			print ("error: reference mismatch!")
			print ("make sure that the contigs of the bam file and the reference match")
			quit()

	if cov.size:
		coverage_norm=numpy.median(cov)
	else:
		coverage_norm=1

	if args.ref:
		coverage_data=gc_norm(args,coverage_norm,chromosomes,coverage_data,Ncontent)

	chromosomal_average=0
	outfile=open(args.o+".ploidy.tab", 'w')
	outfile.write("Contig\tploidy_rounded\tploidy_raw\tmedian_coverage\n")
	for chromosome in chromosomes:
	
		if args.ref:
			N_count=Ncontent[chromosome]
			cov=coverage_data[chromosome][numpy.where( (N_count > -1) & ( (coverage_data[chromosome][:,1] > args.Q) | (coverage_data[chromosome][:,1] == 0) ) ),0]
		else:
			cov=coverage_data[chromosome][numpy.where( (coverage_data[chromosome][:,1] > args.Q) | (coverage_data[chromosome][:,1] == 0) ),0]

		if cov.size:
			chromosomal_average=numpy.median(cov)
		else:
			chromosomal_average=0

		if not args.force_ploidy:
			try:
				ploidies[chromosome]=int(round((chromosomal_average)/coverage_norm*args.n))
			except:
				ploidies[chromosome]=args.n
		else:
			ploidies[chromosome]=args.n  
		library_stats["chr_cov"][chromosome]=chromosomal_average
		
		outfile.write("{}\t{}\t{}\t{}\n".format(chromosome,ploidies[chromosome],round( library_stats["chr_cov"][chromosome]/coverage_norm*args.n,2),library_stats["chr_cov"][chromosome]))

	outfile.close()
	return(ploidies,library_stats,coverage_data)

#normalise the coverage based on GC content
def gc_norm(args,median_coverage,normalising_chromosomes,coverage_data,Ncontent):
	gc_vals=range(0,101)
	gc_dict={}

	for gc in gc_vals:
		tmp=[]
		gc_dict[gc]=median_coverage
		for chromosome in normalising_chromosomes:
			selected= Ncontent[chromosome] == gc
			selected_coverage=coverage_data[chromosome][selected]
			tmp+=list(selected_coverage[ numpy.where( selected_coverage[:,1] > args.q)[0] ,0])
		if len(tmp):
			gc_dict[gc]=numpy.median(tmp)
		if 0 == gc_dict[gc]:
			gc_dict[gc]=median_coverage

	for chromosome in coverage_data:
		for i in range(0,len(coverage_data[chromosome])):
			if not Ncontent[chromosome][i] == -1:
				coverage_data[chromosome][i,0]=median_coverage*coverage_data[chromosome][i,0]/gc_dict[ Ncontent[chromosome][i] ]

	return(coverage_data)

#load the GC tab, and find which bins contain too many N
def retrieve_N_content(args):

	chromosome_size={}
	for line in open(args.o+".gc.wig"):
		if "Per bin GC values" in line:
			continue
		elif "chrom=" in line:
			chromosome=line.split("chrom=")[-1].split()[0]
			continue

		elif "Per bin fraction of N" in line:
			break
		if not chromosome in chromosome_size:
			chromosome_size[chromosome]=0
		chromosome_size[chromosome]+=1
	
	Ncontent={}
	for chromosome in chromosome_size:
		Ncontent[chromosome]=numpy.zeros( chromosome_size[chromosome],dtype=numpy.int8 )
	read_n=False
	for line in open(args.o+".gc.wig"):
		if "Per bin GC values" in line:
			continue
		elif "chrom=" in line:
			chromosome=line.split("chrom=")[-1].split()[0]
			chromosome_index=0
			continue
		elif "Per bin fraction of N" in line:
			read_n=True
			continue

		content=line.strip()
	
		if read_n:
			n=float(content[-1])
			if n > args.n_mask:
				Ncontent[chromosome][chromosome_index]=-1
		else:
			gc=int(round(float(content)*100))
			Ncontent[chromosome][chromosome_index]=gc 

		chromosome_index+=1
	return(Ncontent)



from distutils.core import setup
from Cython.Build import cythonize

setup(
    name = "TIDDIT",
    ext_modules = cythonize(['TIDDIT_calling.py',"TIDDIT_coverage.py","TIDDIT_filtering.py","TIDDIT_signals.py","DBSCAN.py"]),
)



import numpy
import math
import copy
import sys
import sqlite3
import os
import time

import DBSCAN
import TIDDIT_coverage
import TIDDIT_filtering
import TIDDIT_signals

def retrieve_coverage(chromosome,start,end,coverage_data):
	start_index=int(math.floor(start/50.0))
	end_index=int(math.floor(end/50.0))+1
	regional_coverage=coverage_data[chromosome][start_index:end_index,0]
	coverage=numpy.average(regional_coverage[numpy.where(regional_coverage > -1)])
	max_coverage=max(regional_coverage[numpy.where(regional_coverage > -1)])
	quality=numpy.average(coverage_data[chromosome][start_index:end_index,1])
	return (coverage,max_coverage,int(round(quality)))


#Find the number of discordant pairs within a region
def retrieve_discs(chromosome,start,end,coverage_data):
	discs=0
	start_index=int(math.floor(start/50.0))
	end_index=int(math.floor(end/50.0))+1
	discs=sum(coverage_data[chromosome][start_index:end_index,2])
	return(discs)

def count_ref(chromosome,pos,span_data):
	index=int(math.floor(pos/50.0))
	dr=span_data[chromosome][index,0]
	rr=span_data[chromosome][index,1]
	return(dr,rr)

#split inversions into two separate calls
def redefine_inv(vcf_line,signals,library_stats,args):
	analysed_signals=DBSCAN.analyse_pos(signals,True,library_stats,args)
	vcf_line[1]=analysed_signals["posA"]
	vcf_line[7]=vcf_line[7].split(";END=")[0]+";END="+str(analysed_signals["posB"]) +";SVLEN=" + vcf_line[7].split(";SVLEN=")[-1]
	vcf_line[7]=vcf_line[7].split("SVLEN=")[0]+"SVLEN="+str(analysed_signals["posB"]-analysed_signals["posA"]+1) +";COVM=" + vcf_line[7].split(";COVM=")[-1]

	vcf_line[7]=vcf_line[7].split("CIPOS=")[0]+"CIPOS={},{}".format(analysed_signals["min_A"]-analysed_signals["posA"],analysed_signals["max_A"]-analysed_signals["posA"]) + ";CIEND" + vcf_line[7].split(";CIEND")[-1]
	vcf_line[7]=vcf_line[7].split("CIEND=")[0]+"CIEND={},{}".format(analysed_signals["min_B"]-analysed_signals["posB"],analysed_signals["max_B"]-analysed_signals["posB"]) + ";END=" + vcf_line[7].split(";END=")[-1]
	return(vcf_line)

def inv_recluster(vcf_line,candidate,library_stats,args):
	RR=[]
	FF=[]

	for signal in candidate["signals"]:
		if( not signal[2] and not signal[4] ):
			RR.append(signal)
		else:
			FF.append(signal)

	vcf_line_rr=copy.copy(vcf_line)
	vcf_line_ff=copy.copy(vcf_line)
	vcf_line_rr=redefine_inv(vcf_line_rr,numpy.array(RR),library_stats,args)
	vcf_line_ff=redefine_inv(vcf_line_ff,numpy.array(FF),library_stats,args)
	var_id=list(vcf_line_ff[2])
	var_id[-1]="2"
	vcf_line_ff[2]="".join(var_id)

	return([vcf_line_rr,vcf_line_ff])

def generate_vcf_line(chrA,chrB,n,candidate,args,library_stats,percentiles,span_data):
	vcf_line=[]
	if chrA == chrB and  candidate["posA"] > candidate["posB"]:
		candidate["posA"]=candidate["min_A"]
		candidate["posB"]=candidate["max_B"]

	if candidate["discs"] > candidate["splits"]:
		pairs=candidate["discs"]
	else:
		pairs=candidate["splits"]

	percentile_list=numpy.arange(0,101,1)
	p=0
	if candidate["covA"] < candidate["covB"]:
		coverage=candidate["covA"]
		ploidy=library_stats["ploidies"][chrA]
	else:
		coverage=candidate["covB"]
		ploidy=library_stats["ploidies"][chrB]
	i=0
	perc=0
	p=0
	for percentile in percentiles:
		if pairs*ploidy > percentile*coverage:
			p=percentile_list[i]
			perc=percentile
		i+=1
	if p:
		qual=str(p)
	elif not p and not len(percentiles):
		qual="."
	else:
		qual="0"


	a_dr,a_rr=count_ref(chrA,candidate["posA"],span_data)
	b_dr,b_rr=count_ref(chrB,candidate["posB"],span_data)

	split_ratio=0
	if candidate["splits"]:
		split_ratio=candidate["splits"]/float(candidate["splits"]+min([a_rr,b_rr]))
	disc_ratio=0
	if candidate["discs"]:
		disc_ratio=candidate["discs"]/float(candidate["discs"]+min([a_dr,b_dr]))

	candidate["ratio"]=1
	if candidate["splits"] and candidate["splits"] > candidate["discs"]:
		candidate["ratio"]=split_ratio
	elif candidate["discs"]:
		candidate["ratio"]=disc_ratio


	vcf_line.append(chrA)
	vcf_line.append(candidate["posA"])
	vcf_line.append("SV_{}_1".format(n))
	vcf_line.append("N")
	var,variant_type,GT=TIDDIT_filtering.fetch_variant_type(chrA,chrB,candidate,args,library_stats,disc_ratio,split_ratio)
	vcf_line.append(var)
	vcf_line.append(str(qual))
	vcf_line.append(TIDDIT_filtering.fetch_filter(chrA,chrB,candidate,args,library_stats))
	INFO="{};CIPOS={},{};CIEND={},{}".format(variant_type,candidate["min_A"]-candidate["posA"],candidate["max_A"]-candidate["posA"],candidate["min_B"]-candidate["posB"],candidate["max_B"]-candidate["posB"])
	if chrA == chrB:
		if not "BND" in variant_type:
			INFO += ";END={};SVLEN={}".format(candidate["posB"],abs(candidate["posB"]-candidate["posA"]+1))
		INFO += ";COVM={}".format(candidate["covM"])
	stats=";COVA={};COVB={};LFA={};LFB={};LTE={};OR={},{},{},{};ORSR={},{};QUALA={};QUALB={}".format(candidate["covA"],candidate["covB"],int(round(candidate["discsA"])),int(round(candidate["discsB"])),candidate["discs"]+candidate["splits"],candidate["FF"],candidate["RR"] ,candidate["RF"],candidate["FR"],candidate["splitsINV"],candidate["splits"]-candidate["splitsINV"],candidate["QRA"],candidate["QRB"])

	INFO+=stats

	vcf_line.append(INFO)
	FORMAT_FORMAT="GT:CN:DV:RV:DR:RR"
	vcf_line.append(FORMAT_FORMAT)
	CN="."
	if "DEL" in var or "DUP" in var:
		if library_stats["Coverage"]:
			CN=int(round(candidate["covM"]/(library_stats["Coverage"]/args.n)))
		elif library_stats["chr_cov"][chrA]:
			CN=int(round(candidate["covM"]/(library_stats["chr_cov"][chrA]/library_stats["ploidies"][chrA])))
		else:
			CN=0

	if "DEL" in var:
		CN=library_stats["ploidies"][chrA]-CN
		if CN < 0:
			CN=library_stats["ploidies"][chrA]

	FORMAT_STR="{}:{}:{}:{}:{},{}:{},{}".format(GT,CN,candidate["discs"],candidate["splits"],a_dr,b_dr,a_rr,b_rr)
	vcf_line.append(FORMAT_STR)

	if not "BND" in variant_type and not "INV" in variant_type:
		return([vcf_line])
	elif "INV" in variant_type:
		if candidate["FF"] and candidate["RR"] and candidate["RR"] > args.p/2 and candidate["FF"] > args.p/2 and candidate["discs"]:
			return(inv_recluster(vcf_line,candidate,library_stats,args))
		else:
			return([vcf_line])
	else:
		vcf_line_a=copy.copy(vcf_line)
		vcf_line_b=copy.copy(vcf_line)
		inverted=False
		before=True
		if candidate["posA"] == candidate["max_A"]:
			before=False

		if candidate["FF"] + candidate["RR"]  > candidate["RF"] + candidate["FR"]:
			inverted=True

		if not inverted and not before:
			alt_str_a="N[{}:{}[".format(chrB,candidate["posB"])
			alt_str_b="]{}:{}]N".format(chrA,candidate["posA"])
		elif not inverted and before:
			alt_str_a="]{}:{}]N".format(chrB,candidate["posB"])
			alt_str_b="N[{}:{}[".format(chrA,candidate["posA"])
		elif inverted and  not before:
			alt_str_a="N]{}:{}]".format(chrB,candidate["posB"])
			alt_str_b="[{}:{}[N".format(chrA,candidate["posA"])
		else:
			alt_str_a="[{}:{}[N".format(chrB,candidate["posB"])
			alt_str_b="N]{}:{}]".format(chrA,candidate["posA"])

		vcf_line_a[4]=alt_str_a
		vcf_line_b[4]=alt_str_b
		vcf_line_b[2]="SV_{}_2".format(n)
		vcf_line_b[0]=chrB
		vcf_line_b[1]=candidate["posB"]
		return([vcf_line_a,vcf_line_b])

#main function
def cluster(args):
	start_time=time.time()

	if args.ref:
		print ("Loading GC wig file")
		Ncontent=TIDDIT_coverage.retrieve_N_content(args)
	else:
		Ncontent=[]

	print("Loading coverage wig file")
	coverage_data,span_data=TIDDIT_coverage.coverage(args)

	print ("Constructing SV signal database")
	conn = sqlite3.connect(args.o+".db")
	args.c = conn.cursor()

	tableListQuery = "SELECT name FROM sqlite_master WHERE type=\'table\'"
	args.c.execute(tableListQuery)
	tables = map(lambda t: t[0], args.c.fetchall())
	if "TIDDITcall" in tables:
		args.c.execute("DROP TABLE TIDDITcall")

	args.c.execute("CREATE TABLE TIDDITcall (chrA TEXT,chrB TEXT,posA INT,posB INT,forwardA INT,forwardB INT,qualA INT, qualB INT,cigarA TEXT,cigarB TEXT, resolution INT,name INT)")
	header,chromosomes,library_stats=TIDDIT_signals.signals(args,coverage_data)
	args.c.execute("CREATE INDEX CHR ON TIDDITcall (chrA, chrB)")

	print ("Estimating ploidies")
	ploidies,library_stats,coverage_data=TIDDIT_coverage.determine_ploidy(args,chromosomes,coverage_data,Ncontent,library_stats)
	library_stats["ploidies"]=ploidies

	percentiles_disc,percentiles_splits=TIDDIT_signals.sample(args,coverage_data,span_data)

	if not args.e:
		args.e=int(math.sqrt(library_stats["MeanInsertSize"]*2)*12)
	n=1
	print ("clustering signals on chromosome:")
	calls={}
	for chrA in chromosomes:
		calls[chrA] =[]
		print ("{}".format(chrA))
		signals_chrA={}
		signals=[ [hit[0],hit[1],hit[2],hit[3],hit[4],hit[5],hit[6],hit[7],hit[8]] for hit in args.c.execute('SELECT chrB,posA,posB,forwardA,qualA,forwardB,qualB,resolution,name FROM TIDDITcall WHERE chrA == \'{}\''.format(chrA)).fetchall()]

		for chrB in chromosomes:
			if not chrB in signals_chrA:
				signals_chrA[chrB] = []

		for signal in signals:
			signals_chrA[signal[0]].append(signal[1:])

		for chrB in signals_chrA:
			signals_chrA[chrB]=numpy.array(signals_chrA[chrB])
		
		for chrB in chromosomes:
			signal_data=signals_chrA[chrB]
			if not len(signal_data):
				continue

			candidates=DBSCAN.generate_clusters(chrA,chrB,signal_data,library_stats,args)

			for i in range(0,len(candidates)):
				candidates[i]["covA"],candidates[i]["MaxcovA"],candidates[i]["QRA"]=retrieve_coverage(chrA,candidates[i]["min_A"],candidates[i]["max_A"],coverage_data)
				candidates[i]["covB"],candidates[i]["MaxcovB"],candidates[i]["QRB"]=retrieve_coverage(chrB,candidates[i]["min_B"],candidates[i]["max_B"],coverage_data)
				if chrA == chrB:
					if candidates[i]["posB"] > candidates[i]["posA"]:
						candidates[i]["covM"],candidates[i]["MaxcovM"],candidates[i]["QRM"]=retrieve_coverage(chrB,candidates[i]["posA"],candidates[i]["posB"],coverage_data)
					else:
						candidates[i]["covM"],candidates[i]["MaxcovM"],candidates[i]["QRM"]=retrieve_coverage(chrB,candidates[i]["posB"],candidates[i]["posA"],coverage_data)
				else:
					candidates[i]["covM"]=0
					candidates[i]["QRM"]=0
				candidates[i]["discsA"]=retrieve_discs(chrA,candidates[i]["min_A"],candidates[i]["max_A"],coverage_data)
				candidates[i]["discsB"]=retrieve_discs(chrB,candidates[i]["min_B"],candidates[i]["max_B"],coverage_data)
				sizeA=candidates[i]["max_A"]-candidates[i]["min_A"]+library_stats["ReadLength"]
				sizeB=candidates[i]["max_B"]-candidates[i]["min_B"]+library_stats["ReadLength"]
				gap=[]
				for signal in candidates[i]["signals"]:
					gap.append(abs(sizeA-(signal[0]-candidates[i]["min_A"])+(signal[1]-candidates[i]["min_B"])))
				gap=numpy.average(gap)
				if sizeA > library_stats["MeanInsertSize"] or sizeB > library_stats["MeanInsertSize"]:
					gap=1
				elif gap > sizeA:
					gap = gap - sizeA
				else:
					gap=1
				if ploidies[chrA]:
					coverageA=candidates[i]["covA"]/ploidies[chrA]
				else:
					coverageA=candidates[i]["covA"]/float(args.n)
				if ploidies[chrB]:
					coverageB=candidates[i]["covB"]/ploidies[chrB]
				else:
					coverageB=candidates[i]["covB"]/float(args.n)

				if candidates[i]["discs"] > candidates[i]["splits"]:
					vcf_line=generate_vcf_line(chrA,chrB,n,candidates[i],args,library_stats,percentiles_disc,span_data)
				else:
					vcf_line=generate_vcf_line(chrA,chrB,n,candidates[i],args,library_stats,percentiles_splits,span_data)

				if len(vcf_line) == 1:
					calls[chrA].append(vcf_line[0])
				else:
					calls[chrA].append(vcf_line[0])
					if not chrB in calls:
						calls[chrB]=[]
					calls[chrB].append(vcf_line[1])
				n+=1

	outfile=open(args.o+".vcf", 'w')
	new_header=[]
	for line in header.strip().split("\n"):
		if "##TIDDITcmd" in line:
			new_header.append("##TIDDITcmd=\"{}\"\n".format(" ".join(sys.argv)))
			continue
		new_header.append(line+"\n")
	header="".join(new_header)
	outfile.write(header)
	for chromosome in chromosomes:
		for call in sorted(calls[chromosome],key=lambda x: x[1]):
			output=call
			output[1]=str(output[1])
			if "MinSize" == output[6]:
				continue

			output="\t".join(output)+"\n"
			outfile.write(output)

	print ("variant clustering completed in {}".format(time.time()-start_time))
	print ("Work complete!")
	os.remove("{}.db".format(args.o))
	return()




import numpy

#The functions of this script perform DBSCAN and returns the variatn clusters

#Analyse the reads of the cluster, and colelct statistics such as position and orientation
def analyse_pos(candidate_signals,discordants,library_stats,args):
	analysed_signals={}

	max_a=max(candidate_signals[:,0:1])[0]
	min_a=min(candidate_signals[:,0:1])[0]

	max_b=max(candidate_signals[:,1:2])[0]
	min_b=min(candidate_signals[:,1:2])[0]

	
	FF=0
	FR=0
	RF=0
	RR=0

	splitsINV=0
	splits=0

	discs=0

	avgQb=[]
	avgQa=[]

	for i in range(0,len(candidate_signals)):
		if candidate_signals[i][-2] == 1:
			if(candidate_signals[i][2] != candidate_signals[i][4]):
				splitsINV +=1
		else:
			if(candidate_signals[i][2] and not candidate_signals[i][4]):
				FR+=1
			elif( not candidate_signals[i][2] and candidate_signals[i][4]):
				RF+=1
			elif( not candidate_signals[i][2] and not candidate_signals[i][4]):
				RR+=1
			else:
				FF+=1

		if candidate_signals[i][-2] == 1:
			splits +=1
		else:
			discs+=1

		if candidate_signals[i][3] != -1:
			avgQa.append(candidate_signals[i][3])
		if candidate_signals[i][5] != -1:
			avgQb.append(candidate_signals[i][5])

	if avgQa:
		avgQa=numpy.mean(avgQa)
	else:
		avgQa=-1

	if avgQb:
		avgQb=numpy.mean(avgQb)
	else:
		avgQb=-1

	major_orientation=max([FF,FR,RF,RR])
	if not discs:
				posA=max_a
				posB=min_b		
	else:
		if library_stats["Orientation"] == "innie":
			if major_orientation == FF:
				posA=max_a
				posB=max_b
			elif major_orientation == FR:
				posA=max_a
				posB=min_b
			elif major_orientation == RF:
				posA=min_a
				posB=max_b
			else:
				posA=min_a
				posB=min_b
		else:
			if major_orientation == FF:
				posA=min_a
				posB=min_b
			elif major_orientation == FR:
				posA=min_a
				posB=max_b
			elif major_orientation == RF:
				posA=max_a
				posB=min_b
			else:
				posA=max_a
				posB=max_b
	analysed_signals={"posA":posA,"posB":posB,"min_A":min_a,"max_A":max_a,"min_B":min_b,"max_B":max_b,"QA":avgQa,"QB":avgQb,"FF":FF,"FR":FR,"RF":RF,"RR":RR,"splitsINV":splitsINV,"splits":splits,"discs":discs,"signals":candidate_signals}
	return(analysed_signals)

#Call the cluser algorithm, the statistics function, and returns the final cluster
def generate_clusters(chrA,chrB,coordinates,library_stats,args):
	candidates=[]
	coordinates=coordinates[numpy.lexsort((coordinates[:,1],coordinates[:,0]))]
	min_pts=args.l
	if chrA == chrB and library_stats["ploidies"][chrA] > args.n*2:
		min_pts=int(round(args.l/float(args.n)*library_stats["ploidies"][chrA]))

	db=main(coordinates[:,0:2],args.e,min_pts)
	unique_labels = set(db)

	for var in unique_labels:
		if var == -1:
			continue
		class_member_mask = (db == var)
		candidate_signals =coordinates[class_member_mask]
		resolution=candidate_signals[:,-2]
		support=len(set(candidate_signals[:,-1]))
		discordants=True
		if len(set(resolution)) == 1 and max(resolution) == 1:
			disordants=False

		if discordants and support >= args.p:
			candidates.append( analyse_pos(candidate_signals,discordants,library_stats,args) )
		elif not discordants and support >= args.r and chrA == chrB:
			candidates.append( analyse_pos(candidate_signals,discordants,library_stats,args) )

	return(candidates)

def x_coordinate_clustering(data,epsilon,m):
	clusters=numpy.zeros(len(data))
	for i in range(0,len(clusters)):
		clusters[i]=-1
	cluster_id=-1
	cluster=False

	for i in range(0,len(data)-m+1):

		distances=[]
		current=data[i,:]
		points=data[i+1:i+m,:]
		#print points
		distances=[]
		for point in points:
			distances.append(abs(point[0]-current[0]))

		if max(distances) < epsilon:
			#add to the cluster
			if cluster:
				clusters[i+m-1]=cluster_id
				#define a new cluster
			else:
				cluster_id+=1
				cluster=True
				for j in range(i,i+m):
					clusters[j]=cluster_id
		else:
			cluster=False

	return(clusters,cluster_id)

def y_coordinate_clustering(data,epsilon,m,cluster_id,clusters):

	cluster_id_list=set(clusters)
	for cluster in cluster_id_list:
		if cluster == -1:
			continue
		class_member_mask = (clusters == cluster)
		indexes=numpy.where(class_member_mask)[0]
		signals=data[class_member_mask]

		y_coordinates=[]


		for i  in range(0,len(signals)):
			y_coordinates.append([signals[i][1],indexes[i]])
		y_coordinates.sort(key=lambda x:x[0])
		
		sub_clusters=numpy.zeros(len(indexes))
		for i in range(0,len(sub_clusters)):
			sub_clusters[i]=-1

		active_cluster=False
		sub_cluster_id=0
		y_coordinates=numpy.array(y_coordinates)
		for i in range(0,len(y_coordinates)-m+1):
			distances=[]
			current=y_coordinates[i,:]
			next=y_coordinates[i+1:i+m,:]

			distances=[]
			for pos in next:
				distances.append(abs(pos[0]-current[0]))	

			if max(distances) < epsilon:
				#add to the cluster
				if active_cluster:
					sub_clusters[i+m-1]=sub_cluster_id
					#define a new cluster
				else:
					sub_cluster_id+=1
					active_cluster=True
					for j in range(i,i+m):
						sub_clusters[j]=sub_cluster_id
			else:
				active_cluster=False

		for i in range(0,len(sub_clusters)):
			if sub_clusters[i] == 1:
				clusters[ y_coordinates[i][1] ]	= cluster

			elif sub_clusters[i] > -1:
				clusters[ y_coordinates[i][1] ]=sub_clusters[i] +cluster_id-1
			elif sub_clusters[i] == -1:
				clusters[ y_coordinates[i][1] ] = -1

		if sub_cluster_id > 1:
			cluster_id += sub_cluster_id-1
	return(clusters,cluster_id)

def main(data,epsilon,m):
	clusters,cluster_id=x_coordinate_clustering(data,epsilon,m)
	clusters,cluster_id=y_coordinate_clustering(data,epsilon,m,cluster_id,clusters)

	return(clusters)
